#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2023 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Iterator, List, Optional, Set, Tuple
import re

from ..term import graph_to_term
from ..graph import Graph
from ..rewrite import dpo
from ..rule import Rule
from ..matcher import Match, match_rule, find_iso
from .. import state

RULE_NAME_RE = re.compile('(-)?\\s*([a-zA-Z_][\\.a-zA-Z0-9_]*)')

class Tactic:
    __state: state.State
    __local_state: state.RewriteState

    def __init__(self, local_state: state.RewriteState, args: List[str]) -> None:
        self.__local_state = local_state
        self.__state = local_state.state
        self.__goal_lhs: Optional[Graph] = None
        self.__goal_rhs: Optional[Graph] = None
        # self.__goal_stack: List[Tuple[Graph,Graph]] = []
        self.args = args

    def has_goal(self) -> bool:
        return self.__goal_lhs is not None and self.__goal_rhs is not None

    # def push_goal(self) -> None:
    #     if not self.__goal_lhs or not self.__goal_rhs: return
    #     self.__goal_stack.append((self.__goal_lhs.copy(), self.__goal_rhs.copy()))

    # def pop_goal(self) -> bool:
    #     if len(self.__goal_stack) > 0:
    #         self.__goal_lhs, self.__goal_rhs = self.__goal_stack.pop()
    #         return True
    #     else:
    #         return False

    def lookup_rule(self, rule_expr: str) -> Tuple[Optional[Rule],bool]:
        m = RULE_NAME_RE.match(rule_expr)
        if not m: return (None, False)
        converse = m.group(1) == '-'
        rule_name = m.group(2)
        if rule_name in self.__state.rule_sequence and self.__state.rule_sequence[rule_name] <= self.__local_state.sequence:
            rule = self.__state.rules[rule_name]
            if converse:
                return (rule.converse(), True)
            else:
                return (rule.copy(), False)
        else:
            return (None, False)

    def rewrite_lhs(self, rule_expr: str) -> Iterator[Tuple[Match,Match]]:
        if not self.__goal_lhs: return None
        rule, converse = self.lookup_rule(rule_expr)
        if not rule: return None
        if converse and not rule.equiv:
            return None

        for m_g in match_rule(rule, self.__goal_lhs):
            for m_h in dpo(rule, m_g):
                self.__goal_lhs = m_h.cod.copy()
                yield (m_g, m_h)

    def rewrite_rhs(self, rule_expr: str) -> Iterator[Tuple[Match,Match]]:
        if not self.__goal_rhs: return None
        rule, converse = self.lookup_rule(rule_expr)
        if not rule: return None
        if not converse and not rule.equiv:
            return None

        for m_g in match_rule(rule, self.__goal_rhs):
            for m_h in dpo(rule, m_g):
                self.__goal_rhs = m_h.cod.copy()
                yield (m_g, m_h)

    def rewrite_lhs1(self, rule_expr: str) -> bool:
        for _ in self.rewrite_lhs(rule_expr):
            return True
        return False

    def rewrite_rhs1(self, rule_expr: str) -> bool:
        for _ in self.rewrite_rhs(rule_expr):
            return True
        return False

    def validate_goal(self) -> Optional[Match]:
        if not self.__goal_lhs or not self.__goal_rhs: return None

        iso = find_iso(self.__goal_lhs, self.__goal_rhs)
        if (self.__local_state.status != state.RewriteState.INVALID and iso):
            self.__local_state.status = state.RewriteState.VALID
            return iso
        else:
            return None

    def lhs(self) -> Optional[Graph]:
        if not self.__goal_lhs is None:
            return self.__goal_lhs.copy()
        else:
            return None

    def rhs(self) -> Optional[Graph]:
        if not self.__goal_rhs is None:
            return self.__goal_rhs.copy()
        else:
            return None

    def highlight_lhs(self, vertices: Set[int], edges: Set[int]) -> None:
        if self.__local_state.lhs:
            self.__local_state.lhs.highlight(vertices, edges)

    def highlight_rhs(self, vertices: Set[int], edges: Set[int]) -> None:
        if self.__local_state.rhs:
            self.__local_state.rhs.highlight(vertices, edges)

    def __reset(self) -> None:
        self.__goal_lhs = self.__local_state.lhs.copy() if self.__local_state.lhs else None
        self.__goal_rhs = self.__local_state.rhs.copy() if self.__local_state.rhs else None
        # self.__goal_stack = []

    def next_rhs(self, current: str) -> Optional[str]:
        self.__reset()
        seen: Set[str] = set()
        seen = set([current])

        found_prev = (current == '?')
        next_term = None
        for g in self.make_rhs():
            t = graph_to_term(g)
            if found_prev and not t in seen:
                next_term = t
                break
            elif not next_term:
                next_term = t

            seen.add(t)
            found_prev = (current == t)
        return next_term

    def run_check(self) -> None:
        self.__local_state.status = state.RewriteState.CHECKING
        self.__reset()
        self.check()
        if self.__local_state.status != state.RewriteState.VALID:
            self.__local_state.status = state.RewriteState.INVALID

    # tactics should override this method
    def name(self) -> str:
        return 'refl'

    # tactics should override this method
    def check(self) -> None:
        self.validate_goal()

    # tactics that can synthesize an RHS should override this method
    def make_rhs(self) -> Iterator[Graph]:
        lhs = self.lhs()
        if lhs: yield lhs







#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Dict, List, Optional, Tuple

from .tactic import Tactic
from .tactic.simptac import SimpTac
from .tactic.ruletac import RuleTac

from .parser import parse
from .graph import Graph
from .rule import Rule


class RewriteState:
    UNCHECKED = 0
    CHECKING = 1
    VALID = 2
    INVALID = 3

    def __init__(self,
                 sequence: int,
                 state: State,
                 term_pos: Tuple[int,int] = (0,0),
                 equiv: bool=True,
                 tactic: str='',
                 tactic_args: Optional[List[str]]=None,
                 lhs: Optional[Graph]=None,
                 rhs: Optional[Graph]=None,
                 lhs_match: Optional[Graph]=None,
                 rhs_match: Optional[Graph]=None,
                 stub: bool=False) -> None:
        # store an integer representing my current location. This prevents cyclic dependencies of
        # rules used inside of tactics. TODO: this doesn't work quite right for <= rules, since it
        # doesn't store the location of the rule *and* its converse being proved.
        self.sequence = sequence
        self.state = state
        self.status = RewriteState.UNCHECKED
        self.term_pos = term_pos
        self.equiv = equiv
        # self.rule = rule
        self.lhs = lhs
        self.rhs = rhs
        self.lhs_match = lhs_match
        self.rhs_match = rhs_match

        tactic_args = [] if tactic_args is None else tactic_args
        if tactic == 'rule':
            self.tactic: Tactic = RuleTac(self, tactic_args)
        elif tactic == 'simp':
            self.tactic = SimpTac(self, tactic_args)
        else:
            self.tactic = Tactic(self, tactic_args)
        self.stub = stub

    def check(self) -> None:
        self.tactic.run_check()
        # if self.rule and self.lhs and self.rhs:
        #     # check for any constraints on the LHS and RHS first
        #     if self.lhs_match and not find_iso(self.lhs, self.lhs_match):
        #         self.status = RewriteState.INVALID
        #     if self.rhs_match and not find_iso(self.rhs, self.rhs_match):
        #         self.status = RewriteState.INVALID

        #     # if all LHS/RHS constraints are satisfied, try to prove the rule step by rewriting
        #     if self.status == RewriteState.CHECKING:
        #         for m_lhs in match_rule(self.rule, self.lhs):
        #             for m_rhs in dpo(self.rule, m_lhs):
        #                 iso = find_iso(m_rhs.cod, self.rhs)
        #                 if iso:
        #                     self.status = RewriteState.VALID

        #                     for v in m_lhs.dom.vertices():
        #                         self.lhs.vertex_data(m_lhs.vmap[v]).highlight = True
        #                     for e in m_lhs.dom.edges():
        #                         self.lhs.edge_data(m_lhs.emap[e]).highlight = True

        #                     for v in m_rhs.dom.vertices():
        #                         self.rhs.vertex_data(iso.vmap[m_rhs.vmap[v]]).highlight = True
        #                     for e in m_rhs.dom.edges():
        #                         self.rhs.edge_data(iso.emap[m_rhs.emap[e]]).highlight = True

        #                     break

        # if self.status != RewriteState.VALID:
        #     self.status = RewriteState.INVALID

class State:
    def __init__(self) -> None:
        self.graphs: Dict[str, Graph] = dict()
        self.rules: Dict[str, Rule] = dict()
        self.rule_sequence: Dict[str, int] = dict()
        self.rewrites: Dict[str, RewriteState] = dict()
        self.parts: List[Tuple[int, int, str, str]] = list()
        self.errors: List[Tuple[str, int, str]] = list()

    def update(self, code: str, file_name: str) -> None:
        parse_data = parse(code, file_name)
        self.graphs = parse_data.graphs
        self.rules = parse_data.rules
        self.rule_sequence = parse_data.rule_sequence
        self.parts = parse_data.parts
        self.errors = parse_data.errors

        for name, (sequence, t_start, t_end, equiv, tactic, tactic_args, lhs, rhs, lhs_match, rhs_match) in parse_data.rewrites.items():
            stub = not (':' in name)
            self.rewrites[name] = RewriteState(sequence, self, (t_start, t_end), equiv, tactic, tactic_args, lhs, rhs, lhs_match, rhs_match, stub)

    def part_with_index_at(self, pos: int) -> Optional[Tuple[int, Tuple[int,int,str,str]]]:
        p0 = (0, self.parts[0]) if len(self.parts) >= 1 else None
        for (i,p) in enumerate(self.parts):
            if p[0] <= pos:
                p0 = (i,p)
                if p[1] >= pos:
                    return (i,p)
        return p0

    def part_at(self, pos: int) -> Optional[Tuple[int,int,str,str]]:
        p = self.part_with_index_at(pos)
        return p[1] if p else None




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2023 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Iterator

from ..graph import Graph
from . import Tactic


class RuleTac(Tactic):
    def name(self) -> str:
        return 'rule'

    def make_rhs(self) -> Iterator[Graph]:
        if len(self.args) == 0: raise StopIteration()
        for _, m_rhs in self.rewrite_lhs(self.args[0]):
            yield m_rhs.cod

    def check(self) -> None:
        if len(self.args) == 0:
            return

        # apply a single rewrite rule in all possible ways
        for m_lhs, m_rhs in self.rewrite_lhs(self.args[0]):
            # if the LHS and RHS are isomorphic, close the goal...
            iso = self.validate_goal()
            if iso:
                # ...and highlight the part that was rewritten
                rhs_verts = set(iso.vmap[v] for v in m_rhs.vimg)
                rhs_edges = set(iso.emap[e] for e in m_rhs.eimg)
                self.highlight_lhs(m_lhs.vimg, m_lhs.eimg)
                self.highlight_rhs(rhs_verts, rhs_edges)
                return




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2023 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Callable, Iterator


from ..graph import Graph
from . import Tactic


class SimpTac(Tactic):
    MAX_DEPTH = 256

    def name(self) -> str:
        return 'simp'

    def __repeat(self, rw: Callable[[str], bool]):
        got_match = True
        i = 0
        while got_match and i < SimpTac.MAX_DEPTH:
            got_match = False
            for r in self.args:
                # print('rewriting: ' + r)
                while rw(r) and i < SimpTac.MAX_DEPTH:
                    # print('success')
                    got_match = True
                    i += 1

    def make_rhs(self) -> Iterator[Graph]:
        self.__repeat(self.rewrite_lhs1)
        lhs = self.lhs()
        if lhs: yield lhs

    def check(self) -> None:
        self.__repeat(self.rewrite_lhs1)
        self.__repeat(self.rewrite_rhs1)
        self.validate_goal()

